from jsonschema import ValidationError


class TwineException(Exception):
    """All exceptions raised by the twine framework inherit from TwineException"""


class NotImplementedYet(TwineException):
    """Raised when you attempt to use a function whose high-level API is in place, but which is not implemented yet"""


class TwineValueException(TwineException, ValueError):
    """Raised when a python ValueError is appropriate to ensure all errors still also inherit from TwineException"""


class TwineTypeException(TwineException, TypeError):
    """Raised when a python TypeError is appropriate to ensure all errors still also inherit from TwineException"""


class TwineVersionConflict(TwineException):
    """Raised when the (optional) "twined_version" field in the twine file does not match the current installed version of twined"""


# --------------------- Exceptions relating to the twine itself ------------------------


class InvalidTwine(TwineException):
    """Raised when the specified twine is invalid for some reason"""


class InvalidTwineJson(InvalidTwine):
    """Raised when the JSON in the twine file is broken"""


class InvalidTwineContents(InvalidTwine, ValidationError):
    """Raised when the JSON in the twine file is not valid (eg doesn't match twine schema)"""


# --------------------- Exceptions relating to accessing/setting strands ------------------------


class UnknownStrand(TwineException, ValueError):
    """Raised when referencing a strand which is not defined in ALL_STRANDS"""


class StrandNotFound(TwineException, KeyError):
    """Raised when the attempting to access a strand not present in the twine"""


# --------------------- Exceptions relating to missing files/folders ------------------------


class FolderNotFound(TwineException):
    """Raised when a required folder (e.g. <data_dir>/input) cannot be found"""


class CredentialNotFound(TwineException):
    """Raised when a credential specified in the twine file is not present in either the environment or a .env file"""


class TwineFileNotFound(TwineException, FileNotFoundError):
    """Raised when the specified twine file is not present"""


class ConfigurationValuesFileNotFound(TwineException, FileNotFoundError):
    """Raised when attempting to read configuration values from a file that is missing"""


class ConfigurationManifestFileNotFound(TwineException, FileNotFoundError):
    """Raised when a configuration manifest file is required by a twine, but is not present in the input directory"""


class InputManifestFileNotFound(TwineException, FileNotFoundError):
    """Raised when an input manifest file is required by a twine, but is not present in the input directory"""


class InputValuesFileNotFound(TwineException, FileNotFoundError):
    """Raised when attempting to read input values from a file that is missing"""


class OutputManifestFileNotFound(TwineException, FileNotFoundError):
    """Raised when twined checks that output manifest file has been produced, but it is not present in the output directory"""


class OutputValuesFileNotFound(TwineException, FileNotFoundError):
    """Raised when attempting to read output values from a file that is missing"""


# --------------------- Exceptions relating to validation of JSON data (input, output, config values) ------------------


class InvalidSourceKindException(TwineException):
    """Raised when attempting to use the json loader for a disallowed kind"""


class InvalidValues(TwineException):
    """Raised when JSON data (like Config data, Input Values or Output Values) is invalid"""


class InvalidValuesJson(InvalidValues):
    """Raised when the JSON in the file or string is broken so cannot be parsed"""


class InvalidValuesContents(InvalidValues, ValidationError):
    """Raised when the JSON in the file is not valid according to its matching schema."""


# --------------------- Exceptions relating to validation of manifests ------------------------


class InvalidManifest(TwineException):
    """Raised when a manifest loaded from JSON does not pass validation"""


class InvalidManifestJson(InvalidManifest):
    """Raised when the json in the manifest file is broken"""


class InvalidManifestType(InvalidManifest):
    """Raised when user attempts to create a manifest of an invalid type"""


class InvalidManifestContents(InvalidManifest, ValidationError):
    """Raised when the manifest files are missing or do not match tags, sequences, clusters, extensions etc as required"""


# --------------------- Exceptions relating to access of data using the Twine instance ------------------------

# TODO This is related to filtering files from a manifest. Determine whether this belongs here,
#  or whether we should port the filtering code across from the SDK.
class UnexpectedNumberOfResults(TwineException):
    """Raise when searching for a single data file (or a particular number of data files) and the number of results exceeds that expected"""


# --------------------- Maps allowing customised exceptions per-strand (simplifies code elsewhere) ------------------


file_not_found_map = {
    "twine": TwineFileNotFound,
    "configuration_values": ConfigurationValuesFileNotFound,
    "input_values": InputValuesFileNotFound,
    "output_values": OutputValuesFileNotFound,
    "configuration_manifest": ConfigurationManifestFileNotFound,
    "input_manifest": InputManifestFileNotFound,
    "output_manifest": OutputManifestFileNotFound,
}

# TODO Specialised per-strand exceptions to help drill to the root of the issues
invalid_json_map = {
    "twine": InvalidTwineJson,
    "children": InvalidValuesJson,
    "configuration_values": InvalidValuesJson,
    "input_values": InvalidValuesJson,
    "output_values": InvalidValuesJson,
    "configuration_manifest": InvalidManifestJson,
    "input_manifest": InvalidManifestJson,
    "output_manifest": InvalidManifestJson,
}

# TODO Specialised per-strand exceptions to help drill to the root of the issues
invalid_contents_map = {
    "twine": InvalidTwineContents,
    "children": InvalidValuesContents,
    "configuration_values": InvalidValuesContents,
    "input_values": InvalidValuesContents,
    "output_values": InvalidValuesContents,
    "configuration_manifest": InvalidManifestContents,
    "input_manifest": InvalidManifestContents,
    "output_manifest": InvalidManifestContents,
}



from . import exceptions  # noqa: F401
from . import utils  # noqa: F401
from .twine import (  # noqa: F401
    ALL_STRANDS,
    CHILDREN_STRANDS,
    CREDENTIAL_STRANDS,
    MANIFEST_STRANDS,
    SCHEMA_STRANDS,
    Twine,
)



import importlib.util
import json


# Determines whether numpy is available
_numpy_spec = importlib.util.find_spec("numpy")


class TwinedEncoder(json.JSONEncoder):
    """An encoder which will cope with serialising numpy arrays, ndarrays and matrices to JSON (in list form)

    This is designed to work "out of the box" to help people serialise the outputs from twined applications.
    It does not require installation of numpy - it'll work fine if numpy is not present, so can be used in a versatile
    tool in uncertain environments.

    Example use:
    ```
    from twined.utils import TwinedEncoder
    some_json = {"a": np.array([0, 1])}
    json.dumps(some_json, cls=TwinedEncoder)
    ```

    """

    def default(self, obj):
        if _numpy_spec is not None:
            import numpy

            if isinstance(obj, numpy.ndarray) or isinstance(obj, numpy.matrix):
                return obj.tolist()

        return json.JSONEncoder.default(self, obj)



import io
import json
import logging

from twined.exceptions import InvalidSourceKindException


logger = logging.getLogger(__file__)


ALLOWED_KINDS = ("file-like", "filename", "string", "object")


def load_json(source, *args, **kwargs):
    """Loads json, automatically detecting whether the input is a valid filename, a string containing json data,
    or a python dict already (in which case the result is returned directly).

    That makes this function suitable for use in a pipeline where it's not clear whether data has been loaded yet, or
    whether it's in a file or a raw string

    :parameter source: The data source, which can be a string filename ending in *.json (json loaded from disc to
    python dict), a file-like object, a string containing raw json data (json loaded from string to python dict), or
    any other valid python object (passed through).

    :parameter args, kwargs: Arguments passed through to json.load or json.loads, enabling use of custom encoders etc.
    """
    allowed_kinds = kwargs.pop("allowed_kinds", ALLOWED_KINDS)

    def check(kind):
        if kind not in allowed_kinds:
            raise InvalidSourceKindException(f"Attempted to load json from a {kind} data source")

    if isinstance(source, io.IOBase):
        logger.debug("Detected source is a file-like object, loading contents...")
        check("file-like")
        return json.load(source, *args, **kwargs)

    elif not isinstance(source, str):
        logger.debug("Source is not a string, bypassing (returning raw data)")
        check("object")
        return source

    elif source.endswith(".json"):
        logger.debug("Detected source is name of a *.json file, loading from %s", source)
        check("filename")
        with open(source) as f:
            return json.load(f, *args, **kwargs)

    else:
        logger.debug("Detected source is string containing json data, parsing...")
        check("string")
        return json.loads(source, *args, **kwargs)



def trim_suffix(text, suffix):
    """Strip a suffix from text, if it appears (otherwise return text unchanged)"""
    if not text.endswith(suffix):
        return text
    return text[: len(text) - len(suffix)]



from setuptools import find_packages, setup


# Note:
#   The Hitchiker's guide to python provides an excellent, standard, method for creating python packages:
#       http://docs.python-guide.org/en/latest/writing/structure/
#
#   To deploy on PYPI follow the instructions at the bottom of:
#       https://packaging.python.org/tutorials/distributing-packages/#uploading-your-project-to-pypi

with open("README.md") as f:
    readme_text = f.read()

with open("LICENSE") as f:
    license_text = f.read()

setup(
    name="twined",
    version="0.0.17",
    py_modules=[],
    install_requires=["jsonschema ~= 3.2.0", "python-dotenv"],
    url="https://www.github.com/octue/twined",
    license="MIT",
    author="Octue (github: octue)",
    description="A library to help digital twins and data services talk to one another",
    long_description=readme_text,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests", "docs")),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    keywords=["digital", "twins", "data", "services", "python", "schema"],
)
